#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <algorithm>

#include "plugin.hpp"

#include "types.hpp"
#include "types_zxzip.hpp"
#include "manager.hpp"
#include "detector.hpp"
#include "tools.hpp"
#include "tools_zxzip.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Detector*         detector;
extern Options           op;

HANDLE Manager::open(int fNum)
{
  if(fNum >= noVolumes) return INVALID_HANDLE_VALUE;
  
  return CreateFileW(volumeName[fNum],
                    GENERIC_READ,
                    FILE_SHARE_READ | FILE_SHARE_WRITE,
                    NULL,
                    OPEN_EXISTING,
                    FILE_FLAG_SEQUENTIAL_SCAN,
                    NULL);
}

bool Manager::setPointer(DWORD offset)
{
  int vNum = 0;
  for(; vNum < noVolumes; ++vNum)
  {
    if(volumeSize[vNum] < offset)
      offset -= volumeSize[vNum];
    else
      break;
  }
  if(vNum == noVolumes) return false;
  if(vNum == volumeNum)
  {
    SetFilePointer(hostFile, offset + sizeof(HoHdr), NULL, FILE_BEGIN);
    return true;
  }
  CloseHandle(hostFile);

  hostFile = open(vNum);
  if(hostFile == INVALID_HANDLE_VALUE) return false;
  SetFilePointer(hostFile, offset + sizeof(HoHdr), NULL, FILE_BEGIN);
  volumeNum = vNum;
  return true;
}

bool Manager::read(void* buf, WORD size)
{
  DWORD noBytesRead;
  ReadFile(hostFile, buf, size, &noBytesRead, NULL);
  if(noBytesRead == size) return true;
  
  if(++volumeNum == noVolumes) return false;
  
  CloseHandle(hostFile);
  hostFile = open(volumeNum);
  
  if(hostFile == INVALID_HANDLE_VALUE) return false;
  SetFilePointer(hostFile, sizeof(HoHdr), NULL, FILE_BEGIN);
  ReadFile(hostFile, (BYTE*)buf + noBytesRead, size - noBytesRead, &noBytesRead, NULL);

  return true;
}

Manager::Manager(const wchar_t* fileName)
{
  lstrcpyW(hostFileName, fileName);
  noFiles   = 0;
  volumeNum = 0;
  
  ZeroMemory(&lastModifed, sizeof(lastModifed));

  HANDLE hFile = CreateFileW(hostFileName,
                            GENERIC_READ,
                            FILE_SHARE_READ | FILE_SHARE_WRITE,
                            NULL,
                            OPEN_EXISTING,
                            FILE_FLAG_SEQUENTIAL_SCAN,
                            NULL);
  DWORD fileSize = GetFileSize(hFile, NULL);
  CloseHandle(hFile);

  volumeSize  [0] = fileSize - sizeof(HoHdr);
  volumeName  [0] = new wchar_t[300];
  lstrcpyW(volumeName[0], hostFileName);

  wchar_t name[300];
  lstrcpyW(name, hostFileName);
  int len = lstrlenW(name);
  if(name[len-1] != '0')
  {
    len++;
    lstrcatW(name, L"0");
  }
  else
    name[len-1]++;

  for(noVolumes = 1; noVolumes < 10; noVolumes++)
  {
    HANDLE volume = CreateFileW(name,
                               GENERIC_READ,
                               FILE_SHARE_READ | FILE_SHARE_WRITE,
                               NULL,
                               OPEN_EXISTING,
                               FILE_FLAG_SEQUENTIAL_SCAN,
                               NULL);
    if(volume == INVALID_HANDLE_VALUE) break;

    HoHdr hdr;
    DWORD noBytesRead;

    ReadFile(volume, &hdr, sizeof(HoHdr), &noBytesRead, NULL);
    fileSize = GetFileSize(volume, NULL);
    CloseHandle(volume);

    if(hdr.checkSum != calculateCheckSum(hdr)) continue;

    if(memcmp(&hdr.name, "********", 8) || memcmp(&hdr.type, "ZIP", 3)) break;
    volumeSize  [noVolumes] = fileSize - sizeof(HoHdr);
    volumeName  [noVolumes] = new wchar_t[300];
    lstrcpyW(volumeName[noVolumes], name);

    name[len-1]++;
  }
}

Manager::~Manager()
{
  for(int i = 0; i < noVolumes; ++i)
    delete[] volumeName[i];
}

void Manager::GetOpenPanelInfo(struct OpenPanelInfo *info)
{
  info->StructSize    = sizeof(*info);
  info->Flags         = OPIF_ADDDOTS | OPIF_SHOWNAMESONLY | OPIF_SHOWPRESERVECASE;
  info->HostFile      = hostFileName;
  info->CurDir        = L"";
  info->Format        = L"ZXZIP";
  info->StartSortMode = SM_UNSORTED;
  
  static wchar_t panelTitle[260];
  panelTitle[0] = 0;
  lstrcatW(panelTitle, L" ZXZIP:");
  lstrcatW(panelTitle, pointToName(hostFileName));
  lstrcatW(panelTitle, L" ");
  info->PanelTitle = panelTitle;

  static PanelMode   mode[10];
  static const wchar_t*       columnTitles3[4];
  columnTitles3[0] = GetMsg(MName);
  columnTitles3[1] = GetMsg(MSectorSize);
  columnTitles3[2] = GetMsg(MName);
  columnTitles3[3] = GetMsg(MSectorSize);

  static const wchar_t*       columnTitles4[6];
  columnTitles4[0] = GetMsg(MName);
  columnTitles4[1] = GetMsg(MSize);
  columnTitles4[2] = GetMsg(MStart);
  columnTitles4[3] = GetMsg(MSectorSize);
  columnTitles4[4] = GetMsg(MPackedSize);
  columnTitles4[5] = GetMsg(MMethod);

  static const wchar_t*       columnTitles5[2];
  columnTitles5[0] = GetMsg(MName);
  columnTitles5[1] = GetMsg(MFormat);

  static const wchar_t*       columnTitles6[2];
  columnTitles6[0] = GetMsg(MName);
  columnTitles6[1] = GetMsg(MComment);
  
  // C0 - trdos name
  // C1 - trdos 'start'
  // C2 -   
  // C3 - packedSize
  // C4 -  
  // C5 -  
  // C6 -   
  
  mode[3].ColumnTypes=L"N,C2,N,C2";
  mode[3].ColumnWidths=L"0,3,0,3";
  mode[3].ColumnTitles=columnTitles3;
  mode[3].StatusColumnTypes=L"C0,S,C2";
  mode[3].StatusColumnWidths=L"0,5,3";
  mode[3].Flags = PMFLAGS_ALIGNEXTENSIONS;
  
  mode[4].ColumnTypes=L"C0,S,C1,C2,C3,C4";
  mode[4].ColumnWidths=L"0,5,5,3,5,3";
  mode[4].StatusColumnTypes=L"N,S,C2";
  mode[4].StatusColumnWidths=L"0,5,3";
  mode[4].ColumnTitles=columnTitles4;
  
  mode[5].ColumnTypes=L"C0,C5";
  mode[5].ColumnWidths=L"12,0";
  mode[5].StatusColumnTypes=L"N,S,C2";
  mode[5].StatusColumnWidths=L"0,5,3";
  mode[5].ColumnTitles=columnTitles5;

  mode[6].ColumnTypes=L"C0,C6";
  mode[6].ColumnWidths=L"12,0";
  mode[6].StatusColumnTypes=L"N,S,C2";
  mode[6].StatusColumnWidths=L"0,5,3";
  mode[6].ColumnTitles=columnTitles6;
  
  info->PanelModesArray = mode;
  info->PanelModesNumber = sizeof(mode)/sizeof(mode[0]);
  info->StartPanelMode = '0' + op.defaultPanelMode;
}

int Manager::getFindData(PluginPanelItem **pPanelItem, size_t *pNoItems, int opMode)
{
  if(!readInfo()) return FALSE;

  PluginPanelItem *item = new PluginPanelItem[noFiles]{};
  *pNoItems   = noFiles;
  *pPanelItem = item;

  if(noFiles == 0) return TRUE;

  //      trdos' 
  PanelInfo panel = { sizeof(PanelInfo) };
  startupInfo.PanelControl(this, FCTL_GETPANELINFO, 0, &panel);
  int nameColumnWidth = 12;

  if(panel.ViewMode == 4)
  {
    // 25 -   ,  2 -  
    nameColumnWidth = panel.PanelRect.right - panel.PanelRect.left - 2 - 25;
  }
  if(nameColumnWidth < 12) nameColumnWidth = 12;
  
  wchar_t* name = new wchar_t[nameColumnWidth+1];
  
  for(int fNum = 0; fNum < noFiles; ++fNum)
  {
    item[fNum].FileName = pcFiles[fNum].name;
    item[fNum].FileSize = files[fNum].size;
    
    item[fNum].CustomColumnNumber = 7;
    wchar_t **CustomColumnData = new LPWSTR[item[fNum].CustomColumnNumber];
    item[fNum].CustomColumnData = CustomColumnData;
    
    CustomColumnData[0] = new wchar_t[nameColumnWidth+1];

    std::fill_n(name, nameColumnWidth+1, L' ');

    makeTrDosName(name, files[fNum], op.showExt, nameColumnWidth);
    lstrcpyW(CustomColumnData[0], name);
    
    CustomColumnData[1] = new wchar_t[6];
    wsprintfW(CustomColumnData[1], L"%5d", files[fNum].start);

    CustomColumnData[2] = new wchar_t[4];
    wsprintfW(CustomColumnData[2],
             L"%3d",
             files[fNum].noSecs);

    CustomColumnData[3] = new wchar_t[6];
    wsprintfW(CustomColumnData[3], L"%5d", files[fNum].packedSize);

    wchar_t* method = new wchar_t[4];
    switch(files[fNum].pMethod)
    {
      case 0:
                lstrcpyW(method, L"Str");
                break;
      case 1:
                lstrcpyW(method, L"LZP");
                break;
      case 2:
                lstrcpyW(method, L"Shr");
                break;
      case 3:
                if(files[fNum].flags & 0x01)
                 lstrcpyW(method, L"Imp");
                else
                 lstrcpyW(method, L"imp");
                break;
      default:
                lstrcpyW(method, L"???");
    }
    CustomColumnData[4] = method; 

    const wchar_t* description = detector->description(pcFiles[fNum].type);
    if(description)
    {
      CustomColumnData[5] = new wchar_t[lstrlenW(description)+1];
      lstrcpyW(CustomColumnData[5], description);
    }
    else
    {
      CustomColumnData[5] = new wchar_t[1];
      CustomColumnData[5][0] = 0;
    }
    
    const wchar_t* comment = pcFiles[fNum].comment;
    if(comment[0])
    {
      CustomColumnData[6] = new wchar_t[lstrlenW(comment)+1];
      lstrcpyW(CustomColumnData[6], comment);
    }
    else
    {
      CustomColumnData[6] = new wchar_t[1];
      CustomColumnData[6][0] = 0;
    }
  }
  delete[] name;
  return TRUE;
}

void Manager::freeFindData(PluginPanelItem *panelItem, int noItems)
{
  for (int i = 0; i < noItems; i++)
  {
    delete[] panelItem[i].CustomColumnData[0];
    delete[] panelItem[i].CustomColumnData[1];
    delete[] panelItem[i].CustomColumnData[2];
    delete[] panelItem[i].CustomColumnData[3];
    delete[] panelItem[i].CustomColumnData[4];
    delete[] panelItem[i].CustomColumnData[5];
    delete[] panelItem[i].CustomColumnData[6];
    delete[] panelItem[i].CustomColumnData;
  }
  delete[] panelItem;
}

bool Manager::readInfo(void)
{
  //       
  WIN32_FIND_DATAW data;
  HANDLE h = FindFirstFileW(hostFileName, &data);
  if(h == INVALID_HANDLE_VALUE) return false;
  FindClose(h);
  
  if(CompareFileTime(&data.ftLastWriteTime, &lastModifed.ftLastWriteTime) != 0 ||
     data.nFileSizeLow != lastModifed.nFileSizeLow)
  {
    lastModifed = data;
    if(!openHostFile()) return false;

    noFiles = 0;
    DWORD offset = 0;
    for(;;)
    {
      ZipedFileHdr zHdr;
      if(!setPointer(offset))
      {
        errorMessage(GetMsg(MEndOfArchive));
        break;
      }

      if(!read((BYTE*)&zHdr, sizeof(ZipedFileHdr)))
        break;

      if(!isValidZippedFile(zHdr)) break;
      pcFiles[noFiles].offset = offset + sizeof(ZipedFileHdr);
      offset += sizeof(ZipedFileHdr) + zHdr.packedSize;
      CopyMemory(&files[noFiles++], &zHdr, sizeof(ZipedFileHdr));
    }
    makePCNames();
    closeHostFile();
  }
  return true;
}
